
;**************************************************************************
;
; GENERATOR AUTOMATIC THROTTLE SWITCH
;
; (Table-driven throttle curve to come out of idle slowly to prevent the
; generator from stalling)
;
; PIC 12C509 processor drives RC servo to force throttle linkage to idle.
; Servo gets continuous 25mS period signal whose pulse width is either
; 0.6mS (don't force to idle) or 1.8mS (force to idle).
;
; GP0 (pin 7) = current sense input (high when current)
; GP1 (pin 6) = unused
; GP2 (pin 5) = PWM output to servo
; GP3 (pin 4) = MCLR input (low to reset)
; GP4 (pin 3) = low voltage sense input (low when low voltage condition)
; GP5 (pin 2) = adjust servo travel by 1/2 notch on servo crank
;
; Internally: GP0, GP1 input pulled high to 25K, GP3 pulled to 410K.
; Internal 4MHz RC oscillator is used.
;
; --------------------------------|--------------------------
;     CURRENT       LOW VOLTAGE   |     OUTPUT PULSE
;      SENSE           SENSE      |        WIDTH
;       GP0             GP4       |         GP2
; --------------------------------|--------------------------
;  1 (current)      1 (adequate)  |  0.6mS  (allow governor)
;  X                0 (low)       |  0.6mS  (allow governor)
;  0 (no current)   1 (adequate)  |  1.8mS  (pull to idle)
; --------------------------------|--------------------------
;
; Low voltage sense forces the servo to the "allow governor" state so
; that the generator will charge the batteries, and so it will end up
; in the full throttle state for restarting.
;
; S_ADJUST pin, GP5, allows fine adjustment of servo travel by offsetting
; the servo travel by 1/2 of a notch on the servo crank.  This allows
; the idle position of the crank to be pulled exactly over center so that
; the servo current will be minimal during idle.
;
; GP3/MCLR has a brownout circuit to ensure that the processor resets
; properly during low-voltage conditions.
;
; Int RC Osc:        4MHz
; CPU cycle time:    1uS
;
; Actual pulses:     613uS min (allow governored full speed)
;                    1810uS max (pull to idle)
;
;**************************************************************************

    list    p=12c509
    include &ltp12c509.inc&gt
;    #include &ltjonsinc.h&gt
    __config    _CP_OFF & _WDT_OFF & _MCLRE_ON & _IntRC_OSC
    radix   dec

C_SENSE_PIN         equ     0       ;GP0
OUT_PULSE_PIN       equ     2       ;GP2
LV_SENSE_PIN        equ     4       ;GP4
S_ADJUST_PIN        equ     5       ;GP5

COUNT_23MS          equ     165
SECONDS_10          equ     200     ;10-second count @ 50mS/count

holdoff_count       equ     0x10    ;number of 50mS cycles to continue governor mode
table_offset        equ     0x11    ;offset into perentage table
toggle_50ms         equ     0x12    ;doubles 25mS count
icount              equ     0x13    ;counter

        ORG 0x03ff
        movlw   0x40                ;nominal oscillator constant

        ORG 0x0000
        movwf   OSCCAL              ;load RC cal value
        movlw   b'11111011'
        tris    GPIO                ;set GP2 for output, rest are inputs
        movlw   b'10000111'         ;/256 clk prescaler, weak pullups enabled
        option                      ;load option reg
        bcf     GPIO,OUT_PULSE_PIN  ;reset outputs
        clrf    holdoff_count
        clrf    table_offset
        clrf    toggle_50ms

loop25: movlw   COUNT_23MS          ;set up for 23mS (90*256uS)
        movwf   TMR0                ;move to timer register
p25:    movf    TMR0,W              ;get timer count
        btfss   STATUS,Z            ;Skip if it's finally zero
        goto    p25                 ;loop again

        ;DO THE FOLLOWING EVERY 23 MS...

        movlw   0x01
        xorwf   toggle_50ms
                                    ;CHECK FOR LOW VOLTAGE CONDITION
        btfss   GPIO,LV_SENSE_PIN   ;skip if low voltage input has enough voltage
        goto    full                ;assume batteris discharged or generator stopping, run up to full throttle

                                    ;CHECK IF CURRENT IS BEING DRAWN
        btfsc   GPIO,C_SENSE_PIN    ;if current sense input is low, force to idle
        goto    full                ;allow governor at full speed

                                    ;Runs down to idle speed via the
                                    ;two-second table curve, gets here
                                    ;every 25mS.
idle:   movf    holdoff_count,W     ;get 10-second count
        btfsc   STATUS,Z            ;skip if not zero yet
        goto    i_r                 ;branch if ready to begin idle ramp
        movf    toggle_50ms,W       ;get toggle flag
        btfss   STATUS,Z            ;skip if time to decrement idle holdoff count
        goto    i_p1
        decf    holdoff_count       ;decrement count
        goto     i_p1

i_r:    movf    toggle_50ms,W       ;get toggle flag
        btfss   STATUS,Z            ;skip if time to increment
        goto    i_p1

        movlw   39                  ;end of table
        xorwf   table_offset,W
        btfss   STATUS,Z            ;skip if end of table
        incf    table_offset        ;increment offset into table

i_p:    call     p_gen              ;do pulse with percentage in W
        goto    loop25

i_p1:   movf    table_offset,W      ;get offset
        goto    i_p
                                    ;Runs up to governor speed via the
                                    ;two-second table curve, gets here
                                    ;every 25mS.
full:   movlw   SECONDS_10          ;reset idle holdoff count to max
        movwf   holdoff_count

        movf    table_offset,W      ;get table pointer, set flags
        btfsc   STATUS,Z            ;check if zero
        goto    f_p                 ;zero already, don't decrement any further

        movf    toggle_50ms,W       ;get toggle flag
        btfsc   STATUS,Z            ;skip if not time to decrement yet
        decf    table_offset        ;decrement offset into table

f_p:    call     p_gen              ;do pulse with percentage in W
        goto    loop25

;**************************************************************************
                                    ;Enter with table offset in table_offset
p_gen:  bsf     GPIO,OUT_PULSE_PIN  ;start output pulse

        movlw   200                 ;600uS fixed width at beginning of pulse
        movwf   icount
d0:     decfsz  icount
        goto    d0

        btfsc   GPIO,S_ADJUST_PIN   ;skip if adjust pin is low
        goto    d2                  ;don't offset by 80uS
        movlw   26                  ;80uS delay
        movwf   icount
d1:     decfsz  icount
        goto    d1

d2:     movf    table_offset,W      ;get table pointer, set flags
        btfsc   STATUS,Z            ;do variable section if not zero
        goto    dend                ;end pulse now

        call    ramp_table          ;get percentage from table into W

        movwf   icount              ;move percentage to count
d3:     decfsz  icount
        goto    d4                  ;continue with added delay
        goto    dend                ;done with added delay up to 1200uS
d4:     nop                         ;take time
        nop
        nop
        nop
        nop
        nop
        nop
        goto    d3

dend:   bcf     GPIO,OUT_PULSE_PIN  ;end output pulse
        retlw   0

;*************************************************************************
ramp_table:
        addwf   PCL
        retlw   0                 ;First byte of table must be 0 (0% of 1200uS, full speed)
        retlw   10                ;This is a ramp table of 40 percentage
        retlw   20                ;values that takes two seconds to
        retlw   22                ;complete (one value every 50mS).
        retlw   24
        retlw   26
        retlw   28
        retlw   30
        retlw   32
        retlw   34
        retlw   36
        retlw   38
        retlw   40
        retlw   42
        retlw   44
        retlw   46
        retlw   48
        retlw   50
        retlw   52
        retlw   54
        retlw   56
        retlw   58
        retlw   60
        retlw   62
        retlw   64
        retlw   66
        retlw   68
        retlw   70
        retlw   72
        retlw   74
        retlw   76
        retlw   78
        retlw   80
        retlw   82
        retlw   84
        retlw   86
        retlw   88
        retlw   90
        retlw   95
        retlw   100           ;last byte must be 100 (100% of 1200uS, idle speed)

    end

